
/************* REGULAR CONTEXT SEQUENCE MATCHING  ***********************

% File: rules.pl
% Author: Temur Kutsia
% Date: 26.11.2005
% This module implements transformation rules for regular context sequence
  matching algorithm
  
Copyright (c) 2005 Temur Kutsia

*************************************************************************/

:-module(rules,[match/3,
          recompute_match_with_anonymous_variables_instantiated/3,
          op(120,yfy,~>)]).

:-use_module(substitutions).
:-use_module(utilities).


                 /*******************************
                 *    Transformation rules      *
                 *******************************/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Basic transformation rules
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% Success
match([],Acc,Acc).

% Trivial
match([X << X|Tail],Acc,Subst):-!,match(Tail,Acc,Subst).

% Nullary funtions

match([[X|Xs] << X|Tail],Acc,Subst):-
        !,
        match([[X|Xs] << [X]|Tail],Acc,Subst).
match([X << [X]|Tail],Acc,Subst):-
        !,
        match(Tail,Acc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Transformation rules for regular constraints on terms
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% ESRET (Empty sequence in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),seq([]),Flag]|L2],Tail),
        !,
        evaluate_flag_for_seq_var(Flag,0),
        append(L1,L2,L3),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> seq([])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> seq([])],NewXs),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~> seq([])],NewTail),
        match([[X|NewXs] << [X|Ys]|NewTail],NewAcc,Subst).

% CRET (Concatenation in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),conc(R1,R2),Flag]|L2],Tail),
        !,
        evaluate_flag_for_seq_var(Flag,Value),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(SeqVar),TempSeqVar1),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(SeqVar),TempSeqVar2),
        new_flags_for_concatenation(Value,TempSeqVar1,NewFlag1,NewFlag2),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> 
                     seq([TempSeqVar1,TempSeqVar2])],NewAcc),
        instance_of_term_list(Xs,
                              [ctx_or_seq_var(SeqVar) ~> 
                               seq([TempSeqVar1,TempSeqVar2])],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~> 
                               seq([TempSeqVar1,TempSeqVar2])],Tail1),
        NewTail=[[TempSeqVar1,R1,NewFlag1],[TempSeqVar2,R2,NewFlag2]|Tail1],
        match([[X,TempSeqVar1,TempSeqVar2|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

% ChRET (Choice in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),or(R1,R2),Flag]|L2],Tail),
        !,
        select_branch(or(R1,R2),Rselected),
        evaluate_flag_for_seq_var(Flag,Value),
        append([[ctx_or_seq_var(SeqVar),Rselected,Value]|L1],L2,NewTail),
        match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|NewTail],Acc,Subst).

% RRET1 (Repetition in a regular expression for terms)


match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),star(_),Flag]|L2],Tail),
        evaluate_flag_for_seq_var(Flag,0),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> seq([])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> seq([])],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~> seq([])],NewTail),
        match([[X|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

% RRET2 (Repetition in a regular expression for terms)


match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),star(R),_]|L2],Tail),
        !,
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(SeqVar),
                                              TempSeqVar),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     seq([TempSeqVar,ctx_or_seq_var(SeqVar)])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~>
                                  seq([TempSeqVar,ctx_or_seq_var(SeqVar)])],
                              Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~>
                               seq([TempSeqVar,ctx_or_seq_var(SeqVar)])],
                           Tail1),
        NewTail=[[TempSeqVar,R,1],[ctx_or_seq_var(SeqVar),star(R),0]|Tail1],
        match([[X,TempSeqVar,ctx_or_seq_var(SeqVar)|Xs1] << [X|Ys]|NewTail],
              NewAcc,Subst).

% SVRET (Sequence variable in a regular expression for terms)

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),
                    ctx_or_seq_var(AnotherSeqVar),0]|L2],Tail),
        !,
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     ctx_or_seq_var(AnotherSeqVar)],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~>
                                  ctx_or_seq_var(AnotherSeqVar)],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~>
                               ctx_or_seq_var(AnotherSeqVar)],NewTail),
        match([[X,ctx_or_seq_var(AnotherSeqVar)|Xs1] << [X|Ys]|NewTail],
              NewAcc,Subst).

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),
                    ctx_or_seq_var(AnotherSeqVar),1]|L2],Tail),
        !,
        create_unique_temporary_fun_or_indvar(fun_or_ind_var(AnotherSeqVar),
                                              fun_or_ind_var(IndVar)),
        compose(Acc,[ctx_or_seq_var(SeqVar) ~>
                     seq([fun_or_ind_var(IndVar),
                          ctx_or_seq_var(AnotherSeqVar)]),
                     ctx_or_seq_var(AnotherSeqVar) ~>
                     seq([fun_or_ind_var(IndVar),
                          ctx_or_seq_var(AnotherSeqVar)])],
                NewAcc),
        instance_of_term_list(Xs,
                              [ctx_or_seq_var(SeqVar) ~>
                               seq([fun_or_ind_var(IndVar),
                                    ctx_or_seq_var(AnotherSeqVar)]),
                               ctx_or_seq_var(AnotherSeqVar) ~>
                               seq([fun_or_ind_var(IndVar),
                                    ctx_or_seq_var(AnotherSeqVar)])],
                              Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,
                           [fun_or_ind_var(SeqVar) ~>
                            seq([fun_or_ind_var(IndVar),
                                 ctx_or_seq_var(AnotherSeqVar)]),
                            ctx_or_seq_var(AnotherSeqVar) ~>
                            seq([fun_or_ind_var(IndVar),
                                 ctx_or_seq_var(AnotherSeqVar)])],
                           NewTail),
        match([[X,fun_or_ind_var(IndVar),
                ctx_or_seq_var(AnotherSeqVar)|Xs1] <<
               [X|Ys]|NewTail],
              NewAcc,Subst).


% TRET (term in a regular expression for terms)
 
match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(SeqVar),Term,_]|L2],Tail),
        !,
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> Term],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> Term],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(SeqVar) ~> Term],NewTail),
        match([[X,Term|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Transformation rules for regular expressions on contexts
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

% HREC (Hole in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),hole,Flag]|L2],Tail),
        !,
        evaluate_flag_for_ctx_var(Flag,0),
        append(L1,L2,L3),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> hole],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> hole],NewArg),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~> hole],NewTail),
        match([NewArg << T|NewTail],NewAcc,Subst).

% CREC (Concatenation in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),conc(Q1,Q2),Flag]|L2],Tail),
        !,
        evaluate_flag_for_ctx_var(Flag,Value),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxVar),
                                              TempCtxVar1),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxVar),
                                              TempCtxVar2),
        new_flags_for_concatenation(Value,TempCtxVar1,NewFlag1,NewFlag2),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~>
                     [TempCtxVar1,[TempCtxVar2,hole]]],
                NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~>
                      [TempCtxVar1,[TempCtxVar2,hole]]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar)~>
                               [TempCtxVar1,[TempCtxVar2,hole]]],
                           Tail1),
        NewTail=[[TempCtxVar1,Q1,NewFlag1],[TempCtxVar2,Q2,NewFlag2]|Tail1],
        match([[TempCtxVar1,[TempCtxVar2,NewArg]] << T|NewTail],NewAcc,Subst).

% ChREC (Choice in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),or(Q1,Q2),Flag]|L2],Tail),
        !,
        select_branch(or(Q1,Q2),Rselected),
        evaluate_flag_for_ctx_var(Flag,Value),
        append([[ctx_or_seq_var(CtxVar),Rselected,Value]|L1],L2,NewTail),
        match([[ctx_or_seq_var(CtxVar),Arg] << T|NewTail],Acc,Subst).

% RREC1 (Repetition in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),star(_),Flag]|L2],Tail),
        evaluate_flag_for_ctx_var(Flag,0),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> hole],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> hole],NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~> hole],NewTail),
        match([NewArg << T|NewTail],NewAcc,Subst).

% RREC2 (Repetition in a regular expression for contexts)


match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),star(R),_]|L2],Tail),
        !,
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(CtxVar),
                                              TempCtxVar),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~>
                     [TempCtxVar,[ctx_or_seq_var(CtxVar),hole]]],
                NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~>
                      [TempCtxVar,[ctx_or_seq_var(CtxVar),hole]]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~>
                               [TempCtxVar,
                                [ctx_or_seq_var(CtxVar),hole]]],
                           Tail1),
        NewTail=[[TempCtxVar,R,1],
                 [ctx_or_seq_var(CtxVar),star(R),0]|Tail1],
        match([[TempCtxVar,
                [ctx_or_seq_var(CtxVar),NewArg]] << T|NewTail],
              NewAcc,Subst).

% CVREC (Context variable in a regular expression for contexts)

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),
                    [ctx_or_seq_var(AnotherCtxVar),hole],0]|L2],
               Tail),
        !,
        compose(Acc,[ctx_or_seq_var(CtxVar) ~>
                     [ctx_or_seq_var(AnotherCtxVar),hole]],
                NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~>
                      [ctx_or_seq_var(AnotherCtxVar),hole]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~>
                               [ctx_or_seq_var(AnotherCtxVar),hole]],
                           NewTail),
        match([[ctx_or_seq_var(AnotherCtxVar),NewArg] << T|NewTail],
              NewAcc,Subst).

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),
                   [ctx_or_seq_var(AnotherCtxVar),hole],1]|L2],
               Tail),
        !,
        create_unique_temporary_fun_or_indvar(fun_or_ind_var(AnotherCtxVar),
                                              fun_or_ind_var(FunVar)),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(AnotherCtxVar),
                                              ctx_or_seq_var(SeqVar1)),
        create_unique_temporary_ctx_or_seqvar(ctx_or_seq_var(AnotherCtxVar),
                                              ctx_or_seq_var(SeqVar2)),
        compose(Acc,
                [ctx_or_seq_var(CtxVar) ~>
                 [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                  [ctx_or_seq_var(AnotherCtxVar),hole],
                  ctx_or_seq_var(SeqVar2)],
                 ctx_or_seq_var(AnotherCtxVar) ~>
                 [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                  [ctx_or_seq_var(AnotherCtxVar),hole],
                  ctx_or_seq_var(SeqVar2)]],
                NewAcc),
        instance(Arg,
                 [ctx_or_seq_var(CtxVar) ~>
                  [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                   [ctx_or_seq_var(AnotherCtxVar),hole],
                   ctx_or_seq_var(SeqVar2)],
                  ctx_or_seq_var(AnotherCtxVar) ~>
                  [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                   [ctx_or_seq_var(AnotherCtxVar),hole],
                   ctx_or_seq_var(SeqVar2)]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,
                           [ctx_or_seq_var(CtxVar) ~>
                            [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                             [ctx_or_seq_var(AnotherCtxVar),hole],
                             ctx_or_seq_var(SeqVar2)],
                            ctx_or_seq_var(AnotherCtxVar) ~>
                            [fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                             [ctx_or_seq_var(AnotherCtxVar),hole],
                             ctx_or_seq_var(SeqVar2)]],
                           NewTail),
        match([[fun_or_ind_var(FunVar),ctx_or_seq_var(SeqVar1),
                [ctx_or_seq_var(AnotherCtxVar),NewArg],
                ctx_or_seq_var(SeqVar2)] << T|NewTail],
              NewAcc,Subst).

% CRET (Context in a regular expression for contexts)
 
match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_or_seq_var(CtxVar),Ctx,_]|L2],Tail),
        context(Ctx),
        !,
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> Ctx],NewAcc),
        apply_substitution([[ctx_or_seq_var(CtxVar),Arg]<<T],
                           [ctx_or_seq_var(CtxVar) ~> Ctx],NewEquation),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_or_seq_var(CtxVar) ~> Ctx],NewTail),
        append(NewEquation,NewTail,NewProblem),
        match(NewProblem,NewAcc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Rules that do not operate on regular expressions
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% Sequence variable deletion

match([[X,'__'|Xs] << [X|Ys]|Tail],Acc,Subst):-
        match([[X|Xs] << [X|Ys]|Tail],Acc,Subst).

match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> seq([])],NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> seq([])],Xs1),
        apply_substitution(Tail,[ctx_or_seq_var(SeqVar) ~> seq([])],Tail1),
        match([[X|Xs1] << [X|Ys]|Tail1],NewAcc,Subst).

% Widening

match([[X,'__'|Xs] << [X,_|Ys]|Tail],Acc,Subst):-
        !,
        match([[X,'__'|Xs] << [X|Ys]|Tail],Acc,Subst).


match([[X,ctx_or_seq_var(SeqVar)|Xs] << [X,T1|Ys]|Tail],Acc,Subst):-
        !,
        compose(Acc,[ctx_or_seq_var(SeqVar) ~> 
                     seq([T1,ctx_or_seq_var(SeqVar)])],
                NewAcc),
        instance_of_term_list(Xs,[ctx_or_seq_var(SeqVar) ~> 
                                  seq([T1,ctx_or_seq_var(SeqVar)])],
                              Xs1),
        apply_substitution(Tail,[ctx_or_seq_var(SeqVar) ~> 
                                 seq([T1,ctx_or_seq_var(SeqVar)])],
                           Tail1),
        match([[X,ctx_or_seq_var(SeqVar)|Xs1] << [X|Ys]|Tail1],
              NewAcc,Subst).

% Context variable deletion

match([['__',Arg] << T|Tail],Acc,Subst):-
        match([Arg << T|Tail],Acc,Subst).

match([[ctx_or_seq_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        not(member([ctx_or_seq_var(CtxVar),_RegExp,_Flag],Tail)),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> hole],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> hole],Arg1),
        apply_substitution(Tail,[ctx_or_seq_var(CtxVar) ~> hole],Tail1),
        match([Arg1 << T|Tail1],NewAcc,Subst).

% Deepening

match([['__',Arg] << [F|Args]|Tail],Acc,Subst):-
        append([F|_],[Proj|_],[F|Args]),
        match([['__',Arg] << Proj|Tail],Acc,Subst).

match([[ctx_or_seq_var(CtxVar),Arg] << [F|Args]|Tail],Acc,Subst):-
        not(member([ctx_or_seq_var(CtxVar),_RegExp,_Flag],Tail)),
        append([F|L1],[Proj|L2],[F|Args]),
        append([F|L1],[[ctx_or_seq_var(CtxVar),hole]|L2],Context),
        compose(Acc,[ctx_or_seq_var(CtxVar) ~> Context],NewAcc),
        instance(Arg,[ctx_or_seq_var(CtxVar) ~> Context],Arg1),
        apply_substitution(Tail,[ctx_or_seq_var(CtxVar) ~> Context],Tail1),
        match([[ctx_or_seq_var(CtxVar),Arg1] << Proj|Tail1],NewAcc,Subst).

% Decomposition

match([[X,S1|Xs] << [X,T1|Ys]|Tail],Acc,Subst):-!,
        match([S1<<T1,[X|Xs] << [X|Ys]|Tail],Acc,Subst).

% Function variable elimination

match([['_'|Xs] << [Y|Ys]|Tail],Acc,Subst):-
        !,
        match([[Y|Xs] << [Y|Ys]|Tail],Acc,Subst).

match([[fun_or_ind_var(X)|Xs] << [Y|Ys]|Tail],Acc,Subst):-
        !,
        compose(Acc,[fun_or_ind_var(X) ~> Y],NewAcc),
        instance_of_term_list(Xs,[fun_or_ind_var(X) ~> Y],Xs1),
        apply_substitution(Tail,[fun_or_ind_var(X) ~> Y],Tail1),
        match([[Y|Xs1] << [Y|Ys]|Tail1],NewAcc,Subst).
        
match([['_'|Xs] << T|Tail],Acc,Subst):-
        T\=[_|_],
        T\=[],
        !,
        match([[T|Xs] << [T]|Tail],Acc,Subst).

match([[fun_or_ind_var(X)|Xs] << T|Tail],Acc,Subst):-
        T\=[_|_],
        T\=[],
        !,
        compose(Acc,[fun_or_ind_var(X) ~> T],NewAcc),
        instance_of_term_list(Xs,[fun_or_ind_var(X) ~> T],Xs1),
        apply_substitution(Tail,[fun_or_ind_var(X) ~> T],Tail1),
        match([[T|Xs1] << [T]|Tail1],NewAcc,Subst).

% Individual variable elimination

match(['_' << _Term|Tail],Acc,Subst):-
        !,
        match(Tail,Acc,Subst).

match([ fun_or_ind_var(X) << Y|Tail],Acc,Subst):-
        !,
        compose(Acc,[fun_or_ind_var(X) ~> Y],NewAcc),
        apply_substitution(Tail,[fun_or_ind_var(X) ~> Y],Tail1),
        match(Tail1,NewAcc,Subst).
        
match([[_|_]|_],_,_):-
      write('Error. Constrained variable does not occur in matching equations.'),
      nl,abort.


      /*******************************************************************
      *    Recomputing matchers with instantiated  anonymous variables   *
      *******************************************************************/
        
recompute_match_with_anonymous_variables_instantiated(MP,
                                                      OldMatcher,
                                                      NewMatcher):-
        anonymous_to_singleton_variables_in_subst(OldMatcher,
                                                  MatcherWithoutAnonymous),
        recompute_matchers(MP,OldMatcher,MatcherWithoutAnonymous,NewMatcher).


recompute_matchers(_,MatcherWithoutAnonymous,
                   MatcherWithoutAnonymous,MatcherWithoutAnonymous):-!.
recompute_matchers(MP,_,MatcherWithoutAnonymous,NewMatcher):-
        select_equations(MP,Eq),
        apply_substitution(Eq, MatcherWithoutAnonymous,NewMP),
        match(NewMP, MatcherWithoutAnonymous,NewMatcher).


                 /*******************************
                 *    Check for contexts.       *
                 *******************************/


context([hole|Tail]):-!,without_holes(Tail).
context([H|Tail]):-context(H),!,without_holes(Tail).
context([H|Tail]):-without_holes(H),!,context(Tail).

% Check whether a term is without holes, or whether terms in a list
% are all without holes.

without_holes([]):-!.
without_holes([X|Tail]):-without_holes(X),without_holes(Tail).
without_holes(X):-X\=hole.






